@@ -16,32 +16,89 @@
 
 QT_BEGIN_NAMESPACE
 
+/*!
+ * \class QV4::Sequence
+ * \internal
+ *
+ * A Sequence stores the contents of a sequential container and makes them acccessible
+ * to the JavaScript engine. It behaves mostly like a regular JavaScript array. The
+ * entries of the container are exposed as array elements.
+ *
+ * Sequence is a ReferenceObject. Therefore it writes back its contents to the property
+ * it was retrieved from whenever it changes. It also re-reads the property whenever
+ * that one changes.
+ *
+ * As long as a Sequence is attached to a property this way, it is the responsibility of
+ * the property's surrounding (C++) object to keep the contents valid. It has to, for
+ * example, track pointers to QObjects potentially deleted in other places so that they
+ * don't become dangling.
+ *
+ * However, the Sequence can also be detached. This happens predominantly by assigning
+ * it to a QML-declared property. In that case, it becomes the Sequence's responsibility
+ * to track its contents. To do so, it does not necessarily keep an actual instance of
+ * the original container in this case, but may rather stores its contents as actual
+ * JavaScript array elements. This includes QObjectWrappers for all QObject pointers it
+ * may contain. The contents are then marked like all JavaScript array elements when the
+ * garbage collector runs, and QObjectWrapper also guards against external deletion.
+ * There is no property to read or write back in this case, and neither does the
+ * internal container need to be updated. Therefore, the objects stored as array elements
+ * are created detached and won't read or write back.
+ *
+ * For element types that don't need to be marked, Sequence will still use the original
+ * container for storage, even in the detached case. This is usually more efficient
+ * because because it saves some data conversion. The only types that need to be marked
+ * are pointers to QObject-derived types, either stored as-is or hidden inside QVariant.
+ * Whenever the container cannot possibly hold such elements (directly or indirectly),
+ * the original container is used.
+ */
+
 Q_STATIC_LOGGING_CATEGORY(lcListValueConversion, "qt.qml.listvalueconversion")
 
 namespace QV4 {
 
 DEFINE_OBJECT_VTABLE(Sequence);
 
-static ReturnedValue doGetIndexed(const Sequence *s, qsizetype index) {
-    QV4::Scope scope(s->engine());
+static ReturnedValue doGetIndexed(Heap::Sequence *p, qsizetype index)
+{
+    Q_ASSERT(p->isStoredInline());
+    QV4::Scope scope(p->internalClass->engine);
 
-    Heap::ReferenceObject::Flags flags =
-            Heap::ReferenceObject::EnforcesLocation;
-    if (s->d()->metaSequence().canSetValueAtIndex())
+    const QMetaType valueMetaType = p->valueMetaType();
+    const QMetaSequence metaSequence = p->metaSequence();
+
+    Heap::ReferenceObject::Flags flags = Heap::ReferenceObject::EnforcesLocation;
+    if (metaSequence.canSetValueAtIndex())
         flags |= Heap::ReferenceObject::CanWriteBack;
-    if (s->d()->valueMetaType() == QMetaType::fromType<QVariant>())
-        flags |= Heap::ReferenceObject::IsVariant;
 
-    QV4::ScopedValue v(scope, scope.engine->fromVariant(
-                           s->at(index), s->d(), index, flags));
+    const void *container = p->storagePointer();
+    Q_ASSERT(container); // Must readReference() before
+
+    QVariant result;
+    if (valueMetaType == QMetaType::fromType<QVariant>()) {
+        flags |= Heap::ReferenceObject::IsVariant;
+        metaSequence.valueAtIndex(container, index, &result);
+    } else {
+        result = QVariant(valueMetaType);
+        metaSequence.valueAtIndex(container, index, result.data());
+    }
+
+    QV4::ScopedValue v(scope, scope.engine->fromVariant(result, p, index, flags));
     if (QQmlValueTypeWrapper *ref = v->as<QQmlValueTypeWrapper>()) {
         if (CppStackFrame *frame = scope.engine->currentStackFrame)
             ref->d()->setLocation(frame->v4Function, frame->statementNumber());
-        // No need to read the reference. at() has done that already.
+        // No need to read the reference. We've done that above already.
     }
     return v->asReturnedValue();
 }
 
+static void *createVariantData(QMetaType type, QVariant *variant)
+{
+    if (type == QMetaType::fromType<QVariant>())
+        return variant;
+    *variant = QVariant(type);
+    return variant->data();
+}
+
 // helper function to generate valid warnings if errors occur during sequence operations.
 static void generateWarning(QV4::ExecutionEngine *v4, const QString& description)
 {
@@ -58,24 +115,38 @@ static void generateWarning(QV4::ExecutionEngine *v4, const QString& description
     QQmlEnginePrivate::warning(engine, retn);
 }
 
+static qsizetype sizeInline(const Heap::Sequence *p)
+{
+    Q_ASSERT(p->isStoredInline());
+
+    if (const void *container = p->storagePointer())
+        return p->metaSequence().size(container);
+
+    // It can be stored inline, and the container can still be nullptr.
+    // This happens if we construct it from a nullptr in the first place and never update it.
+    // It means it's empty.
+    return 0;
+}
+
 struct SequenceOwnPropertyKeyIterator : ObjectOwnPropertyKeyIterator
 {
     ~SequenceOwnPropertyKeyIterator() override = default;
     PropertyKey next(const Object *o, Property *pd = nullptr, PropertyAttributes *attrs = nullptr) override
     {
-        const Sequence *s = static_cast<const Sequence *>(o);
+        Heap::Sequence *p = static_cast<const Sequence *>(o)->d();
+        Q_ASSERT(p->isStoredInline());
 
-        if (s->d()->isReference() && !s->loadReference())
+        if (p->isReference() && !p->loadReference())
             return PropertyKey::invalid();
 
-        const qsizetype size = s->size();
+        const qsizetype size = sizeInline(p);
         if (size > 0 && qIsAtMostSizetypeLimit(arrayIndex, size - 1)) {
             const uint index = arrayIndex;
             ++arrayIndex;
             if (attrs)
                 *attrs = QV4::Attr_Data;
             if (pd)
-                pd->value = doGetIndexed(s, index);
+                pd->value = doGetIndexed(p, index);
             return PropertyKey::fromArrayIndex(index);
         }
 
@@ -95,48 +166,129 @@ void Heap::Sequence::initTypes(QMetaType listType, QMetaSequence metaSequence)
     Q_ASSERT(m_listType);
     m_metaSequence = metaSequence.iface();
     Q_ASSERT(m_metaSequence);
-    QV4::Scope scope(internalClass->engine);
-    QV4::Scoped<QV4::Sequence> o(scope, this);
-    o->setArrayType(Heap::ArrayData::Custom);
 }
 
 void Heap::Sequence::init(QMetaType listType, QMetaSequence metaSequence, const void *container)
 {
     ReferenceObject::init(nullptr, -1, NoFlag);
     initTypes(listType,  metaSequence);
-    m_container = listType.create(container);
+    if (isStoredInline())
+        createInlineStorage(container);
+    else
+        createElementWrappers(container);
 }
 
 void Heap::Sequence::init(
     QMetaType listType, QMetaSequence metaSequence, const void *container,
     Heap::Object *object, int propertyIndex, Heap::ReferenceObject::Flags flags)
 {
-    ReferenceObject::init(object, propertyIndex, flags);
+    ReferenceObject::init(object, propertyIndex, flags | IsDirty);
     initTypes(listType, metaSequence);
 
-    if (CppStackFrame *frame = internalClass->engine->currentStackFrame)
-        setLocation(frame->v4Function, frame->statementNumber());
-    if (container)
-        m_container = listType.create(container);
-    else if (flags & EnforcesLocation)
-        QV4::ReferenceObject::readReference(this);
+    if (isStoredInline()) {
+        if (CppStackFrame *frame = internalClass->engine->currentStackFrame)
+            setLocation(frame->v4Function, frame->statementNumber());
+        createInlineStorage(container);
+        if (!container && (flags & EnforcesLocation))
+            QV4::ReferenceObject::readReference(this);
+    } else {
+        createElementWrappers(container);
+    }
 }
 
-Heap::Sequence *Heap::Sequence::detached() const
+void Heap::Sequence::createInlineStorage(const void *container)
 {
-    return internalClass->engine->memoryManager->allocate<QV4::Sequence>(
-        QMetaType(m_listType), QMetaSequence(m_metaSequence), m_container);
+    Q_ASSERT(isStoredInline());
+
+    QV4::Scope scope(internalClass->engine);
+    QV4::Scoped<QV4::Sequence> o(scope, this);
+    o->setArrayType(Heap::ArrayData::Custom);
+    if (container)
+        m_container = listType().create(container);
+}
+
+void Heap::Sequence::createElementWrappers(const void *container)
+{
+    Q_ASSERT(!isStoredInline());
+
+    if (!container)
+        return;
+
+    const QMetaSequence metaSequence(m_metaSequence);
+    const QMetaType valueMetaType = metaSequence.valueMetaType();
+    const qsizetype size = metaSequence.size(container);
+
+    QV4::Scope scope(internalClass->engine);
+    if (!qIsAtMostUintLimit(size)) {
+        generateWarning(scope.engine, QLatin1String("Sequence length out of range"));
+        return;
+    }
+
+    QV4::Scoped<QV4::Sequence> self(scope, this);
+    self->arrayReserve(size);
+    QV4::ScopedValue v(scope);
+    if (valueMetaType == QMetaType::fromType<QVariant>()) {
+        QVariant var;
+        for (qsizetype i = 0; i < size; ++i) {
+            metaSequence.valueAtIndex(container, i, &var);
+            v = scope.engine->metaTypeToJS(var.metaType(), var.constData());
+            self->arraySet(i, v);
+        }
+    } else {
+        QVariant var(valueMetaType);
+        for (qsizetype i = 0; i < size; ++i) {
+            metaSequence.valueAtIndex(container, i, var.data());
+            v = scope.engine->metaTypeToJS(valueMetaType, var.constData());
+            self->arraySet(i, v);
+        }
+    }
+
+    m_size = size;
+}
+
+Heap::Sequence *Heap::Sequence::detached()
+{
+    const QMetaType listType(m_listType);
+    const QMetaSequence metaSequence(m_metaSequence);
+    if (isStoredInline()) {
+        return internalClass->engine->memoryManager->allocate<QV4::Sequence>(
+                listType, metaSequence, m_container);
+    }
+
+    QVariant list(listType);
+
+    const QMetaType valueMetaType(m_metaSequence->valueMetaType);
+    QVariant element;
+    void *elementData = createVariantData(valueMetaType, &element);
+
+    QV4::Scope scope(internalClass->engine);
+    if (qIsAtMostSizetypeLimit(m_size)) {
+        QV4::Scoped<QV4::Sequence> self(scope, this);
+        QV4::ScopedValue v(scope);
+        for (uint i = 0; i < m_size; ++i) {
+            v = self->get(PropertyKey::fromArrayIndex(i));
+            ExecutionEngine::metaTypeFromJS(v, valueMetaType, elementData);
+            metaSequence.addValue(list.data(), elementData);
+        }
+    } else {
+        generateWarning(scope.engine, QLatin1String("Index out of range during toVariant()"));
+    }
+
+    return scope.engine->memoryManager->allocate<QV4::Sequence>(
+            listType, metaSequence, list.constData());
 }
 
 void Heap::Sequence::destroy()
 {
-    if (m_container)
+    if (isStoredInline() && m_container)
         listType().destroy(m_container);
     ReferenceObject::destroy();
 }
 
 void *Heap::Sequence::storagePointer()
 {
+    if (!isStoredInline())
+        return nullptr;
     if (!m_container)
         m_container = listType().create();
     return m_container;
@@ -144,6 +296,9 @@ void *Heap::Sequence::storagePointer()
 
 bool Heap::Sequence::setVariant(const QVariant &variant)
 {
+    // Should only happen from readReference(). Therefore we are attached.
+    Q_ASSERT(isStoredInline());
+
     const QMetaType variantReferenceType = variant.metaType();
     if (variantReferenceType != listType()) {
         // This is a stale reference. That is, the property has been
@@ -172,68 +327,12 @@ bool Heap::Sequence::setVariant(const QVariant &variant)
 }
 QVariant Heap::Sequence::toVariant() const
 {
+    // Should only happen from readReference(). Therefore we are attached.
+    Q_ASSERT(isStoredInline());
+
     return QVariant(listType(), m_container);
 }
 
-qsizetype Sequence::size() const
-{
-    const auto *p = d();
-    Q_ASSERT(p->storagePointer()); // Must readReference() before
-    return p->metaSequence().size(p->storagePointer());
-}
-
-QVariant Sequence::at(qsizetype index) const
-{
-    const auto *p = d();
-    Q_ASSERT(p->storagePointer()); // Must readReference() before
-    const QMetaType v = p->valueMetaType();
-    QVariant result;
-    if (v == QMetaType::fromType<QVariant>()) {
-        p->metaSequence().valueAtIndex(p->storagePointer(), index, &result);
-    } else {
-        result = QVariant(v);
-        p->metaSequence().valueAtIndex(p->storagePointer(), index, result.data());
-    }
-    return result;
-}
-
-QVariant Sequence::shift()
-{
-    auto *p = d();
-    void *storage = p->storagePointer();
-    Q_ASSERT(storage); // Must readReference() before
-    const QMetaType v = p->valueMetaType();
-    const QMetaSequence m = p->metaSequence();
-
-    const auto variantData = [&](QVariant *variant) -> void *{
-        if (v == QMetaType::fromType<QVariant>())
-            return variant;
-
-        *variant = QVariant(v);
-        return variant->data();
-    };
-
-    QVariant result;
-    void *resultData = variantData(&result);
-    m.valueAtIndex(storage, 0, resultData);
-
-    if (m.canRemoveValueAtBegin()) {
-        m.removeValueAtBegin(storage);
-        return result;
-    }
-
-    QVariant t;
-    void *tData = variantData(&t);
-    for (qsizetype i = 1, end = m.size(storage); i < end; ++i) {
-        m.valueAtIndex(storage, i, tData);
-        m.setValueAtIndex(storage, i - 1, tData);
-    }
-    m.removeValueAtEnd(storage);
-
-    return result;
-}
-
-
 template<typename Action>
 void convertAndDo(const QVariant &item, const QMetaType v, Action action)
 {
@@ -249,246 +348,251 @@ void convertAndDo(const QVariant &item, const QMetaType v, Action action)
     }
 }
 
-void Sequence::append(const QVariant &item)
+static void appendInline(Heap::Sequence *p, const QVariant &item)
 {
-    Heap::Sequence *p = d();
     convertAndDo(item, p->valueMetaType(), [p](const void *data) {
         p->metaSequence().addValueAtEnd(p->storagePointer(), data);
     });
 }
 
-void Sequence::append(qsizetype num, const QVariant &item)
+static void appendDefaultConstructedInline(Heap::Sequence *p, qsizetype num)
 {
-    Heap::Sequence *p = d();
-    convertAndDo(item, p->valueMetaType(), [p, num](const void *data) {
-        const QMetaSequence m = p->metaSequence();
-        void *container = p->storagePointer();
-        for (qsizetype i = 0; i < num; ++i)
-            m.addValueAtEnd(container, data);
-    });
+    QVariant item;
+    const void *data = createVariantData(p->valueMetaType(), &item);
+    const QMetaSequence m = p->metaSequence();
+    void *container = p->storagePointer();
+    for (qsizetype i = 0; i < num; ++i)
+        m.addValueAtEnd(container, data);
 }
 
-void Sequence::replace(qsizetype index, const QVariant &item)
+static void replaceInline(Heap::Sequence *p, qsizetype index, const QVariant &item)
 {
-    Heap::Sequence *p = d();
     convertAndDo(item, p->valueMetaType(), [p, index](const void *data) {
         p->metaSequence().setValueAtIndex(p->storagePointer(), index, data);
     });
 }
 
-void Sequence::removeLast(qsizetype num)
+static void removeLastInline(Heap::Sequence *p, qsizetype num)
 {
-    auto *p = d();
     const QMetaSequence m = p->metaSequence();
+    void *container = p->storagePointer();
 
     if (m.canEraseRangeAtIterator() && m.hasRandomAccessIterator() && num > 1) {
-        void *i = m.end(p->storagePointer());
+        void *i = m.end(container);
         m.advanceIterator(i, -num);
-        void *j = m.end(p->storagePointer());
-        m.eraseRangeAtIterator(p->storagePointer(), i, j);
+        void *j = m.end(container);
+        m.eraseRangeAtIterator(container, i, j);
         m.destroyIterator(i);
         m.destroyIterator(j);
     } else {
         for (int i = 0; i < num; ++i)
-            m.removeValueAtEnd(p->storagePointer());
+            m.removeValueAtEnd(container);
     }
 }
 
-ReturnedValue Sequence::containerGetIndexed(qsizetype index, bool *hasProperty) const
+bool Heap::Sequence::loadReference()
 {
-    if (d()->isReference() && !loadReference())
+    Q_ASSERT(isReference());
+    // If locations are enforced we only read once
+    return enforcesLocation() || QV4::ReferenceObject::readReference(this);
+}
+
+bool Heap::Sequence::storeReference()
+{
+    Q_ASSERT(isReference());
+    return isAttachedToProperty() && QV4::ReferenceObject::writeBack(this);
+}
+
+ReturnedValue Sequence::virtualGet(const Managed *that, PropertyKey id, const Value *receiver, bool *hasProperty)
+{
+    Heap::Sequence *p = static_cast<const Sequence *>(that)->d();
+    if (!p->isStoredInline() || !id.isArrayIndex())
+        return ReferenceObject::virtualGet(that, id, receiver, hasProperty);
+
+    const uint arrayIndex = id.asArrayIndex();
+    if (!qIsAtMostSizetypeLimit(arrayIndex)) {
+        generateWarning(that->engine(), QLatin1String("Index out of range during indexed get"));
+        return false;
+    }
+
+    if (p->isReference() && !p->loadReference())
         return Encode::undefined();
 
-    if (index >= 0 && index < size()) {
+    const qsizetype index = arrayIndex;
+    if (index < sizeInline(p)) {
         if (hasProperty)
             *hasProperty = true;
-        return doGetIndexed(this, index);
+        return doGetIndexed(p, index);
     }
+
     if (hasProperty)
         *hasProperty = false;
     return Encode::undefined();
 }
 
-bool Sequence::containerPutIndexed(qsizetype index, const Value &value)
-{
-    if (internalClass()->engine->hasException)
-        return false;
-
-    if (d()->isReadOnly()) {
-        engine()->throwTypeError(QLatin1String("Cannot insert into a readonly container"));
-        return false;
-    }
-
-    if (d()->isReference() && !loadReference())
-        return false;
-
-    const qsizetype count = size();
-    const QMetaType valueType = d()->valueMetaType();
-    const QVariant element = ExecutionEngine::toVariant(value, valueType, false);
-
-    if (index < 0)
-        return false;
-
-    if (index == count) {
-        append(element);
-    } else if (index < count) {
-        replace(index, element);
-    } else {
-        /* according to ECMA262r3 we need to insert */
-        /* the value at the given index, increasing length to index+1. */
-        append(index - count,
-               valueType == QMetaType::fromType<QVariant>() ? QVariant() : QVariant(valueType));
-        append(element);
-    }
-
-    if (d()->object())
-        storeReference();
-    return true;
-}
-
-SequenceOwnPropertyKeyIterator *containerOwnPropertyKeys(const Object *m, Value *target)
-{
-    *target = *m;
-    return new SequenceOwnPropertyKeyIterator;
-}
-
-bool Sequence::containerDeleteIndexedProperty(qsizetype index)
-{
-    if (d()->isReadOnly())
-        return false;
-    if (d()->isReference() && !loadReference())
-        return false;
-    if (index < 0 || index >= size())
-        return false;
-
-    /* according to ECMA262r3 it should be Undefined, */
-    /* but we cannot, so we insert a default-value instead. */
-    replace(index, QVariant());
-
-    if (d()->object())
-        storeReference();
-
-    return true;
-}
-
-bool Sequence::containerIsEqualTo(Managed *other)
-{
-    if (!other)
-        return false;
-    Sequence *otherSequence = other->as<Sequence>();
-    if (!otherSequence)
-        return false;
-    if (d()->object() && otherSequence->d()->object()) {
-        return d()->object() == otherSequence->d()->object()
-                && d()->property() == otherSequence->d()->property();
-    } else if (!d()->object() && !otherSequence->d()->object()) {
-        return this == otherSequence;
-    }
-    return false;
-}
-
-void *Sequence::getRawContainerPtr() const
-{ return d()->storagePointer(); }
-
-bool Sequence::loadReference() const
-{
-    Q_ASSERT(d()->object());
-    // If locations are enforced we only read once
-    return d()->enforcesLocation() || QV4::ReferenceObject::readReference(d());
-}
-
-bool Sequence::storeReference()
-{
-    Q_ASSERT(d()->object());
-    return d()->isAttachedToProperty() && QV4::ReferenceObject::writeBack(d());
-}
-
-ReturnedValue Sequence::virtualGet(const Managed *that, PropertyKey id, const Value *receiver, bool *hasProperty)
-{
-    if (id.isArrayIndex()) {
-        const uint index = id.asArrayIndex();
-        if (qIsAtMostSizetypeLimit(index))
-            return static_cast<const Sequence *>(that)->containerGetIndexed(qsizetype(index), hasProperty);
-
-        generateWarning(that->engine(), QLatin1String("Index out of range during indexed get"));
-        return false;
-    }
-
-    return Object::virtualGet(that, id, receiver, hasProperty);
-}
-
 qint64 Sequence::virtualGetLength(const Managed *m)
 {
-    const Sequence *s = static_cast<const Sequence *>(m);
-    if (s->d()->isReference() && !s->loadReference())
+    Heap::Sequence *p = static_cast<const Sequence *>(m)->d();
+    if (!p->isStoredInline())
+        return p->m_size;
+    if (p->isReference() && !p->loadReference())
         return 0;
-    return s->size();
+    return sizeInline(p);
 }
 
 bool Sequence::virtualPut(Managed *that, PropertyKey id, const Value &value, Value *receiver)
 {
-    if (id.isArrayIndex()) {
-        const uint index = id.asArrayIndex();
-        if (qIsAtMostSizetypeLimit(index))
-            return static_cast<Sequence *>(that)->containerPutIndexed(qsizetype(index), value);
+    if (!id.isArrayIndex())
+        return ReferenceObject::virtualPut(that, id, value, receiver);
 
+    const uint arrayIndex = id.asArrayIndex();
+    Sequence *s = static_cast<Sequence *>(that);
+    Heap::Sequence *p = s->d();
+
+    if (!p->isStoredInline()) {
+        s->arraySet(arrayIndex, value);
+        if (p->m_size <= arrayIndex)
+            p->m_size = arrayIndex + 1;
+        return true;
+    }
+
+    if (!qIsAtMostSizetypeLimit(arrayIndex)) {
         generateWarning(that->engine(), QLatin1String("Index out of range during indexed set"));
         return false;
     }
-    return Object::virtualPut(that, id, value, receiver);
+
+    if (p->internalClass->engine->hasException)
+        return false;
+
+    if (p->isReadOnly()) {
+        p->internalClass->engine->throwTypeError(
+                QLatin1String("Cannot insert into a readonly container"));
+        return false;
+    }
+
+    if (p->isReference() && !p->loadReference())
+        return false;
+
+    const qsizetype index = arrayIndex;
+    const qsizetype count = sizeInline(p);
+    const QMetaType valueType = p->valueMetaType();
+    const QVariant element = ExecutionEngine::toVariant(value, valueType, false);
+
+    if (index == count) {
+        appendInline(p, element);
+    } else if (index < count) {
+        replaceInline(p, index, element);
+    } else {
+        /* according to ECMA262r3 we need to insert */
+        /* the value at the given index, increasing length to index+1. */
+        appendDefaultConstructedInline(p, index - count);
+        appendInline(p, element);
+    }
+
+    if (p->isReference())
+        p->storeReference();
+    return true;
 }
 
 bool Sequence::virtualDeleteProperty(Managed *that, PropertyKey id)
 {
-    if (id.isArrayIndex()) {
-        const uint index = id.asArrayIndex();
-        if (qIsAtMostSizetypeLimit(index))
-            return static_cast<Sequence *>(that)->containerDeleteIndexedProperty(qsizetype(index));
+    Heap::Sequence *p = static_cast<const Sequence *>(that)->d();
+    if (!p->isStoredInline() || !id.isArrayIndex())
+        return ReferenceObject::virtualDeleteProperty(that, id);
 
+    const uint arrayIndex = id.asArrayIndex();
+    if (!qIsAtMostSizetypeLimit(arrayIndex)) {
         generateWarning(that->engine(), QLatin1String("Index out of range during indexed delete"));
         return false;
     }
-    return Object::virtualDeleteProperty(that, id);
+
+    if (p->isReadOnly())
+        return false;
+    if (p->isReference() && !p->loadReference())
+        return false;
+
+    const qsizetype index = arrayIndex;
+    if (index >= sizeInline(p))
+        return false;
+
+    /* according to ECMA262r3 it should be Undefined, */
+    /* but we cannot, so we insert a default-value instead. */
+    replaceInline(p, index, QVariant());
+
+    if (p->isReference())
+        p->storeReference();
+
+    return true;
 }
 
 bool Sequence::virtualIsEqualTo(Managed *that, Managed *other)
 {
-    return static_cast<Sequence *>(that)->containerIsEqualTo(other);
+    if (!other)
+        return false;
+
+    const Sequence *otherS = other->as<Sequence>();
+    if (!otherS)
+        return false;
+
+    const Sequence *s = static_cast<Sequence *>(that);
+    const Heap::Sequence *p = s->d();
+    const Heap::Sequence *otherP = otherS->d();
+
+    const Heap::Object *object = p->object();
+    const Heap::Object *otherObject = otherP->object();
+
+    if (object && otherObject)
+        return object == otherObject && p->property() == otherP->property();
+
+    if (!object && !otherObject)
+        return s == otherS;
+
+    return false;
 }
 
 OwnPropertyKeyIterator *Sequence::virtualOwnPropertyKeys(const Object *m, Value *target)
 {
-    return containerOwnPropertyKeys(m, target);
+    Heap::Sequence *p = static_cast<const Sequence *>(m)->d();
+    if (!p->isStoredInline())
+        return ReferenceObject::virtualOwnPropertyKeys(m, target);
+
+    *target = *m;
+    return new SequenceOwnPropertyKeyIterator;
 }
 
 int Sequence::virtualMetacall(Object *object, QMetaObject::Call call, int index, void **a)
 {
-    Sequence *sequence = static_cast<Sequence *>(object);
-    Q_ASSERT(sequence);
+    Heap::Sequence *p = static_cast<Sequence *>(object)->d();
+    Q_ASSERT(p);
+
+    // We only create attached wrappers if this sequence is stored inline.
+    // When detaching, we re-create everything. Therefore, we can't get a metaCall if
+    // we aren't stored inline.
+    Q_ASSERT(p->isStoredInline());
 
     switch (call) {
     case QMetaObject::ReadProperty: {
-        const QMetaType valueType = sequence->d()->valueMetaType();
-        if (sequence->d()->isReference() && !sequence->loadReference())
+        const QMetaType valueType = p->valueMetaType();
+        if (p->isReference() && !p->loadReference())
             return 0;
-        const QMetaSequence metaSequence = sequence->d()->metaSequence();
+        const QMetaSequence metaSequence = p->metaSequence();
         if (metaSequence.valueMetaType() != valueType)
             return 0; // value metatype is not what the caller expects anymore.
 
-        const void *storagePointer = sequence->d()->storagePointer();
+        const void *storagePointer = p->storagePointer();
         if (index < 0 || index >= metaSequence.size(storagePointer))
             return 0;
         metaSequence.valueAtIndex(storagePointer, index, a[0]);
         break;
     }
     case QMetaObject::WriteProperty: {
-        void *storagePointer = sequence->d()->storagePointer();
-        const QMetaSequence metaSequence = sequence->d()->metaSequence();
+        void *storagePointer = p->storagePointer();
+        const QMetaSequence metaSequence = p->metaSequence();
         if (index < 0 || index >= metaSequence.size(storagePointer))
             return 0;
         metaSequence.setValueAtIndex(storagePointer, index, a[0]);
-        if (sequence->d()->isReference())
-            sequence->storeReference();
+        if (p->isReference())
+            p->storeReference();
         break;
     }
     default:
@@ -498,66 +602,85 @@ int Sequence::virtualMetacall(Object *object, QMetaObject::Call call, int index,
     return -1;
 }
 
-static QV4::ReturnedValue method_get_length(const FunctionObject *b, const Value *thisObject, const Value *, int)
+QV4::ReturnedValue SequencePrototype::method_getLength(
+        const FunctionObject *b, const Value *thisObject, const Value *, int)
 {
     QV4::Scope scope(b);
     QV4::Scoped<Sequence> This(scope, thisObject->as<Sequence>());
     if (!This)
         THROW_TYPE_ERROR();
 
-    if (This->d()->isReference() && !This->loadReference())
+    Heap::Sequence *p = This->d();
+    if (!p->isStoredInline())
+        RETURN_RESULT(p->m_size);
+
+    if (p->isReference() && !p->loadReference())
         return Encode::undefined();
 
-    const qsizetype size = This->size();
+    const qsizetype size = sizeInline(p);
     if (qIsAtMostUintLimit(size))
         RETURN_RESULT(Encode(uint(size)));
 
     return scope.engine->throwRangeError(QLatin1String("Sequence length out of range"));
 }
 
-static QV4::ReturnedValue method_set_length(const FunctionObject *f, const Value *thisObject, const Value *argv, int argc)
+QV4::ReturnedValue SequencePrototype::method_setLength(
+        const FunctionObject *f, const Value *thisObject, const Value *argv, int argc)
 {
     QV4::Scope scope(f);
     QV4::Scoped<Sequence> This(scope, thisObject->as<Sequence>());
     if (!This)
         THROW_TYPE_ERROR();
 
+    Heap::Sequence *p = This->d();
+
     bool ok = false;
     const quint32 argv0 = argc ? argv[0].asArrayLength(&ok) : 0;
-    if (!ok || !qIsAtMostSizetypeLimit(argv0)) {
+    if (!ok) {
         generateWarning(scope.engine, QLatin1String("Index out of range during length set"));
         RETURN_UNDEFINED();
     }
 
-    if (This->d()->isReadOnly())
+    if (!p->isStoredInline()) {
+        if (argv0 < p->m_size)
+            p->arrayData->vtable()->truncate(This, argv0);
+        p->m_size = argv0;
+        RETURN_UNDEFINED();
+    }
+
+    if (!qIsAtMostSizetypeLimit(argv0)) {
+        generateWarning(scope.engine, QLatin1String("Sequence length out of range"));
+        RETURN_UNDEFINED();
+    }
+
+    if (p->isReadOnly())
         THROW_TYPE_ERROR();
 
     const qsizetype newCount = qsizetype(argv0);
 
     /* Read the sequence from the QObject property if we're a reference */
-    if (This->d()->isReference() && !This->loadReference())
+    if (p->isReference() && !p->loadReference())
         RETURN_UNDEFINED();
 
     /* Determine whether we need to modify the sequence */
-    const qsizetype count = This->size();
+    const qsizetype count = sizeInline(p);
     if (newCount == count) {
         RETURN_UNDEFINED();
     } else if (newCount > count) {
-        const QMetaType valueMetaType = This->d()->valueMetaType();
         /* according to ECMA262r3 we need to insert */
         /* undefined values increasing length to newLength. */
         /* We cannot, so we insert default-values instead. */
-        This->append(newCount - count, QVariant(valueMetaType));
+        appendDefaultConstructedInline(p, newCount - count);
     } else {
         /* according to ECMA262r3 we need to remove */
         /* elements until the sequence is the required length. */
         Q_ASSERT(newCount < count);
-        This->removeLast(count - newCount);
+        removeLastInline(p, count - newCount);
     }
 
     /* write back if required. */
-    if (This->d()->object())
-        This->storeReference();
+    if (p->isReference())
+        p->storeReference();
 
     RETURN_UNDEFINED();
 }
@@ -565,7 +688,7 @@ static QV4::ReturnedValue method_set_length(const FunctionObject *f, const Value
 void SequencePrototype::init()
 {
     defineDefaultProperty(engine()->id_valueOf(), method_valueOf, 0);
-    defineAccessorProperty(QStringLiteral("length"), method_get_length, method_set_length);
+    defineAccessorProperty(QStringLiteral("length"), method_getLength, method_setLength);
     defineDefaultProperty(QStringLiteral("shift"), method_shift, 0);
 }
 
@@ -582,19 +705,43 @@ ReturnedValue SequencePrototype::method_shift(
     if (!s)
         return ArrayPrototype::method_shift(b, thisObject, argv, argc);
 
-    if (s->d()->isReference() && !s->loadReference())
+    Heap::Sequence *p = s->d();
+
+    if (!p->isStoredInline())
+        return ArrayPrototype::method_shift(b, thisObject, argv, argc);
+
+    if (p->isReference() && !p->loadReference())
         RETURN_UNDEFINED();
 
-    const qsizetype len = s->size();
+    const qsizetype len = sizeInline(p);
     if (!len)
         RETURN_UNDEFINED();
 
-    ScopedValue result(scope, scope.engine->fromVariant(s->shift()));
+    void *storage = p->storagePointer();
+    Q_ASSERT(storage); // Must readReference() before
+    const QMetaType v = p->valueMetaType();
+    const QMetaSequence m = p->metaSequence();
 
-    if (s->d()->object())
-        s->storeReference();
+    QVariant shifted;
+    void *resultData = createVariantData(v, &shifted);
+    m.valueAtIndex(storage, 0, resultData);
 
-    return result->asReturnedValue();
+    if (m.canRemoveValueAtBegin()) {
+        m.removeValueAtBegin(storage);
+    } else {
+        QVariant t;
+        void *tData = createVariantData(v, &t);
+        for (qsizetype i = 1, end = m.size(storage); i < end; ++i) {
+            m.valueAtIndex(storage, i, tData);
+            m.setValueAtIndex(storage, i - 1, tData);
+        }
+        m.removeValueAtEnd(storage);
+    }
+
+    if (p->isReference())
+        p->storeReference();
+
+    return scope.engine->fromVariant(shifted);
 }
 
 ReturnedValue SequencePrototype::newSequence(
@@ -645,18 +792,25 @@ ReturnedValue SequencePrototype::fromData(
 QVariant SequencePrototype::toVariant(const Sequence *object)
 {
     Q_ASSERT(object->isV4SequenceType());
-    const auto *p = object->d();
+    Heap::Sequence *p = object->d();
+
+    if (p->isStoredInline()) {
+        // Note: For historical reasons, we ignore the result of loadReference()
+        //       here. This allows us to retain sequences whose objects have vaninshed
+        //       as "var" properties. It comes at the price of potentially returning
+        //       outdated data. This is the behavior sequences have always shown.
+        if (p->isReference())
+            p->loadReference();
+
+        if (const void *storage = p->m_container)
+            return QVariant(p->listType(), storage);
 
-    // Note: For historical reasons, we ignore the result of loadReference()
-    //       here. This allows us to retain sequences whose objects have vaninshed
-    //       as "var" properties. It comes at the price of potentially returning
-    //       outdated data. This is the behavior sequences have always shown.
-    if (p->isReference())
-        object->loadReference();
-    if (!p->hasData())
         return QVariant();
+    }
 
-    return QVariant(p->listType(), p->storagePointer());
+    QV4::Scope scope(p->internalClass->engine);
+    QV4::ScopedObject q(scope, p);
+    return toVariant(q, p->listType());
 }
 
 bool convertToIterable(QMetaType metaType, void *data, QV4::Object *sequence)
@@ -791,11 +945,86 @@ QVariant SequencePrototype::toVariant(const QV4::Value &array, QMetaType targetT
 
 }
 
-void *SequencePrototype::getRawContainerPtr(const Sequence *object, QMetaType typeHint)
+static Heap::Sequence *resolveHeapSequence(const Sequence *sequence, QMetaType typeHint)
 {
-    if (object->d()->listType() == typeHint)
-        return object->getRawContainerPtr();
-    return nullptr;
+    if (!sequence)
+        return nullptr;
+    Heap::Sequence *p = sequence->d();
+    if (p->listType() != typeHint)
+        return nullptr;
+    return p;
+}
+
+void *SequencePrototype::rawContainerPtr(const Sequence *sequence, QMetaType typeHint)
+{
+    Heap::Sequence *p = resolveHeapSequence(sequence, typeHint);
+    if (!p || !p->isStoredInline())
+        return nullptr;
+
+    return p->storagePointer();
+}
+
+SequencePrototype::RawCopyResult SequencePrototype::setRawContainer(
+        Sequence *sequence, const void *container, QMetaType typeHint)
+{
+    Heap::Sequence *p = resolveHeapSequence(sequence, typeHint);
+    if (!p)
+        return TypeMismatch;
+
+    if (p->isStoredInline()) {
+        if (typeHint.equals(p->m_container, container))
+            return WasEqual;
+        typeHint.destruct(p->m_container);
+        typeHint.construct(p->storagePointer(), container);
+        return Copied;
+    }
+
+    p->createElementWrappers(container);
+    return Copied;
+}
+
+SequencePrototype::RawCopyResult SequencePrototype::getRawContainer(
+        const Sequence *sequence, void *container, QMetaType typeHint)
+{
+    Heap::Sequence *p = resolveHeapSequence(sequence, typeHint);
+    if (!p)
+        return TypeMismatch;
+
+    if (p->isStoredInline()) {
+        if (typeHint.equals(p->m_container, container))
+            return WasEqual;
+        typeHint.destruct(container);
+        typeHint.construct(container, p->m_container);
+        return Copied;
+    }
+
+    typeHint.destruct(container);
+    typeHint.construct(container);
+
+    const QMetaSequence metaSequence = p->metaSequence();
+    QV4::Scope scope(p->internalClass->engine);
+    QV4::ScopedValue val(scope);
+    const QMetaType metaType = p->valueMetaType();
+    const qsizetype size = p->m_size;
+
+    Q_ASSERT(qIsAtMostUintLimit(size));
+
+    if (metaType == QMetaType::fromType<QVariant>()) {
+        for (qsizetype i = 0; i < size; ++i) {
+            val = sequence->get(PropertyKey::fromArrayIndex(i));
+            QVariant var = ExecutionEngine::toVariant(val, QMetaType(), false);
+            metaSequence.addValueAtEnd(container, &var);
+        }
+        return Copied;
+    }
+
+    QVariant var(metaType);
+    for (qsizetype i = 0; i < size; ++i) {
+        val = sequence->get(PropertyKey::fromArrayIndex(i));
+        scope.engine->metaTypeFromJS(val, metaType, var.data());
+        metaSequence.addValueAtEnd(container, var.data());
+    }
+    return Copied;
 }
 
 QMetaType SequencePrototype::metaTypeForSequence(const Sequence *object)
