@@ -116,8 +116,14 @@ static inline QString imageId(const QUrl &url)
 
 QQuickDefaultTextureFactory::QQuickDefaultTextureFactory(const QImage &image)
 {
-    if (image.format() == QImage::Format_ARGB32_Premultiplied
-            || image.format() == QImage::Format_RGB32) {
+    if (image.format() == QImage::Format_ARGB32_Premultiplied ||
+        image.format() == QImage::Format_RGB32 ||
+        image.format() == QImage::Format_RGBA16FPx4_Premultiplied ||
+        image.format() == QImage::Format_RGBA16FPx4 ||
+        image.format() == QImage::Format_RGBX16FPx4 ||
+        image.format() == QImage::Format_RGBA32FPx4_Premultiplied ||
+        image.format() == QImage::Format_RGBA32FPx4 ||
+        image.format() == QImage::Format_RGBX32FPx4) {
         im = image;
     } else {
         im = image.convertToFormat(QImage::Format_ARGB32_Premultiplied);
@@ -437,7 +443,7 @@ QNetworkAccessManager *QQuickPixmapReader::networkAccessManager()
 {
     if (!accessManager) {
         Q_ASSERT(readerThreadExecutionEnforcer());
-        accessManager = QQmlEnginePrivate::get(engine)->createNetworkAccessManager(
+        accessManager = QQmlEnginePrivate::get(engine)->typeLoader.createNetworkAccessManager(
                 readerThreadExecutionEnforcer());
     }
     return accessManager;
@@ -469,6 +475,18 @@ static void maybeRemoveAlpha(QImage *image)
 
             *image = image->convertToFormat(QImage::Format_RGB30);
             break;
+        case QImage::Format_RGBA16FPx4:
+            if (image->data_ptr()->convertInPlace(QImage::Format_RGBX16FPx4, Qt::AutoColor))
+                break;
+
+            *image = image->convertToFormat(QImage::Format_RGBX16FPx4);
+            break;
+        case QImage::Format_RGBA32FPx4:
+            if (image->data_ptr()->convertInPlace(QImage::Format_RGBX32FPx4, Qt::AutoColor))
+                break;
+
+            *image = image->convertToFormat(QImage::Format_RGBX32FPx4);
+            break;
         default:
             if (image->data_ptr()->convertInPlace(QImage::Format_RGB32, Qt::AutoColor))
                 break;
